
/*
Lister engine control sketch.  Initially, this just turns on the water pump at a temperature, if
that's not enough, we then turn on the fan.  As things cool, fan off first, then pump.
Later I will add a PID control for the injector/throttle, and perhaps auto-start and kill
commands for this.  This will slave to a pi in the battery box (ethernet!), and also monitor main battery volts 
and net amperes (and if I get fancy, integrate amp-seconds and estimate state of charge).
The uno just gets raw data and executes commands from pi, not too much autonomous here other than
cooling and speed control stuff.  The pi is better suited for heavy math lifting and major decision making.


 Demostrate the slave protocol for the homestead arduinos, driven by a host serial input,
 with wired-or output from each UNO.
 
 Command format:
 First byte is unit address (we're using all ascii so far, plenty for this for now)
 Second byte is command
 Further bytes if required for command data, up to 32 total bytes including
 '\n' at end which terminates the command sequence.

 Minor error detection/ignore implemented.  Might need more...(so far, works great)
 Protocol stategy is just to have host try again if it doesn't get an answer.
 
 Serial monitor adds required '/n'
 if set to "newline" in the monitor window (which isn't the default for it)
 (This may or may not work with SPI going....or have to pick a different led pin)
 */

#include <Wire.h>
#include <Adafruit_ADS1015.h>
#include <SPI.h>
#include <Adafruit_MAX31855.h>
#include <OneWire.h>
#include <DallasTemperature.h>

#define ONE_WIRE_BUS 3  /* Data wire is plugged into port 3 on the Arduino */
Adafruit_ADS1015 ads;     /* Use this for the 12-bit version */ 
OneWire oneWire(ONE_WIRE_BUS);
DallasTemperature sensors(&oneWire);  
#define CS   10    // for thermocouple amp/breakout board chip select
// rest of spi pins: 11 mosi, 12 miso, 13 sclk
// probably can't use pin 11 for other things when hardware SPI is used - would have been SPI-out
// thermocouple breakout board - a C++ object

Adafruit_MAX31855 thermocouple(CS); // note, MAX31855? lib changed to use CLOCK_DIV32
 
#define FAN_PIN 9  // high true output turns on fan
#define PUMP_PIN 8 // high true output turns on pump

// tune these for cooling system
#define FAN_ON_TEMP 190
#define FAN_OFF_PUMP_ON 187
#define PUMP_OFF_TEMP 185
 
// Pin 13 has an LED connected on this Arduino board (as well, we're connecting one to 8 on the shield).
// give it a name:
#define LED_PIN 8  // gosh, save a byte (it's also fan)
// Pass our oneWire reference to Dallas Temperature. 

#define BAUD_RATE 115200 // pick yer poison

// stuff for command state machine below

#define MY_ADDRESS '1'  // could be any byte, but we're using all-ascii for now, and '\n' for end of command
#define CMD_BUF_SIZE 32 // in theory, two of these would fit in the uno buffer at once.  Probably don't need this much IRL
#define MAX_TIME 20 // milliseconds to timeout if incomplete command (3ms would do for 115200 and fast commands)
byte Command[CMD_BUF_SIZE]; // command buffer for incoming commands
byte CommandIndex = 0; // where am I in above buffer
byte CommandState = 0; // state variable
unsigned long CommandTime; // time at start of command recieve, for timeout error detection
////// done with command-slave state machine stuff

//byte FirstTime; // in case we need this after reset
//unsigned long LastPIDTime; // last PID loop time
//#define PIDLOOPTIME 50    // milliseconds per loop


#define TEMP_TIME 5000  // give things time to respond
#define AD_TIME 100    // interval for volt/amp readings
#define VOLT_MULTIPLIER (.01495F)
#define AMP_MULTIPLIER .25
#define AMP_OFFSET (-.373)
//=  -.294
unsigned long LastTempTime;
unsigned long LastADTime;

#define FILTERI (.01)
#define FILTERA (1-FILTERI)

float volts;
float ampsi; // input before filter
float amps;
float ampseconds;
float listtemp;
float listrpm;
float listcjtemp;
float listamps;
float battemp;
float TcCj; // cold junction temperature
float TcTmp; // thermocouple temperature

//////////////////////////////////////////////// 
void soft_reset () // software reset, actually a hard RST
{ // so we can reset timers and so on under host control
 asm volatile ("  jmp 0"); 
} // this really is a hard reset, but coming from software

////////////////////////////
void Report()
{ // send whatever we have back to host over serial
 Serial.print (volts);
 Serial.print ("\t");
 Serial.print (amps);
 Serial.print ("\t");
 Serial.print (ampseconds);
 Serial.print ("\t");
 Serial.print (battemp);
 Serial.print ("\t");
 Serial.print (TcCj);
 Serial.print ("\t");
 Serial.print (TcTmp);
 Serial.print ("\t");
 Serial.print (listamps);
 Serial.print ("\t");
 Serial.print (digitalRead(PUMP_PIN));
 Serial.print ("\t");
 Serial.print (digitalRead(FAN_PIN));
 Serial.print ("\t");
 Serial.print (listrpm);
 Serial.println(); 
 ampseconds = 0;
}

//////////////
void CommandMachine()
{ // State machine for slaves to a master commander.  If a command is incomprehensible, not for our unit, 
  // or in error, we ignore it, and if it takes too long to complete, we time out and ignore that too.  
  // So the master, if it doesn't get the expected response, is supposed to re-try.  KISS.
 switch (CommandState)
 {
  case 0: // looking for work to do
  if (Serial.available())
  {
   CommandTime = millis(); // for later possible timeout
   CommandState = 1; // get bytes till done or timeout
  }
  break;

  case 1: // getting command
  while (Serial.available())
  {
   Command[CommandIndex] = Serial.read(); // get a byte if available
//   Serial.print ("received:"); // debug
//   Serial.println(Command[CommandIndex]); // debug
   
   if (Command[CommandIndex] == '\n') // single quotes, else compiler thinks pointer to string
    { // got a full command, go parse it
     CommandState = 2;
     return; // get it on next roundy round of loop() 
    }
   CommandIndex++; // bump the position
   if ((CommandIndex >= CMD_BUF_SIZE) || ((CommandTime + MAX_TIME) < millis())) // enough parens?
   { // error of some kind, reset state machine
    CommandIndex = 0;
    CommandState = 0; // ignore what might have been garbage
//    Serial.println ("Command error"); // debug
   }
  } // end while
  break;
  
  case 2: // parse command
   if (Command[0] != MY_ADDRESS) // is this even for us?
   { // no, ditch it
    CommandIndex = 0;
    CommandState = 0;
    return; // forget it ever happened 
   }
   // It might have been for us, any legality tests might go after this comment.
   // By convention, address of unit is first byte, command is next, data might follow, should be enough.
   // For time consuming or big stuff, use a subroutine call instead of inlining it like this example.
   // Commands are all-ascii for convienience of host (and we slaves).
   switch (Command[1])
    {
     case 'a': // host requests an ack
        Serial.println("ack");  // should work
     break;
     
//     case 'l': // for test, led on
//      digitalWrite(LED_PIN, HIGH);   // turn the LED on (HIGH is the voltage level)
//     break; // end 'l' command
//     case 'o': // led off command
//      digitalWrite(LED_PIN, LOW);    // turn the LED off by making the voltage LOW
//     break; // end off

    case 'r': // report our "stuff"
     Report();
    break;
    
    case '!': // halt and catch fire
    soft_reset();
    break;  // unreachable code ;~)
    }    
  // we are done with this one   
     CommandIndex = 0;
     CommandState = 0;
//     Serial.println("end state 2"); // debug
    
   break; // end state 2
 } // end switch state machine
} // end command state machine routine

////////////////////////////////////////////////
void readTC()
{  // get these in farenheit
  TcCj = thermocouple.readInternal() * 1.8 + 32;
  TcTmp = thermocouple.readCelsius() * 1.8 + 32; 
}

/**********************************************************/
// the setup routine runs once when you press reset:
void setup() {                
  // initialize things
  pinMode(LED_PIN, OUTPUT); // various pins into in/out state we want
  pinMode(FAN_PIN, OUTPUT);
  pinMode(PUMP_PIN, OUTPUT);
  digitalWrite(FAN_PIN, LOW); // start with stuff turned off
  digitalWrite(PUMP_PIN, LOW);
  //  FirstTime = 1; // first time flag  
  Serial.begin(BAUD_RATE); // someplace to get commands from
  ads.setGain(GAIN_EIGHT);      // 8x gain   +/- 0.512V  1 bit = 0.25mV   0.015625mV
  ads.begin(); // start a/d lib
  sensors.setWaitForConversion(false); // pipelined
  sensors.setResolution(12);
  sensors.begin(); // start dallas one wire temp sensor
//  Serial.print("devices: ");
//  Serial.println (sensors.getDeviceCount());
  battemp = sensors.requestTemperaturesByIndex(0); // might happen in time, who knows.
  battemp = 0; // because we're one sample behind - start pipline with a defined number
  listamps = listrpm = 0; // till I implement these
  LastADTime = LastTempTime = millis(); // time at start, never quite zero.  Pedantry!
}
/**********************************************************/
// the loop routine runs over and over again forever:
void loop() {
// Cooperative multitasking opsys...back to the future (win 3.1!)
// here we try to do the most critical timing stuff first, then the rest next time around
// if they both hit on the same milli
  CommandMachine(); // look for work to do from host, be responsive, runs every time around loop()
  if (LastADTime + AD_TIME <= millis())
  { // get battery params
   volts = ads.readADC_Differential_2_3() * VOLT_MULTIPLIER;
   ampsi = ads.readADC_Differential_0_1()  * AMP_MULTIPLIER + AMP_OFFSET;
   amps = ampsi * FILTERI + amps * FILTERA; // lowpass filter, amps are really noisy
   ampseconds += amps * 0.1; // tenth second readings
// add lister amps on built in a/d here?
   LastADTime += AD_TIME;
  } else if (LastTempTime+TEMP_TIME <= millis()) // rolls over just like millis() so still correct  
  { // do other stuff between PID loops (the else - spread the love around cycle-wise)
   readTC(); // find out how hot it is, from max 31855 thermocouple breakout
   // kind of "if you weren't there you couldn't hear me anyway", and if you were doing the right thing already,
   // no effect.  Saves a little fooling around - simple.
   if (TcTmp > FAN_ON_TEMP) digitalWrite (FAN_PIN, HIGH); // too hot, fan on
   if (TcTmp > FAN_OFF_PUMP_ON) digitalWrite (PUMP_PIN, HIGH);  // try pump first
   if (TcTmp < FAN_OFF_PUMP_ON) digitalWrite (FAN_PIN, LOW); // too cool for fan
   if (TcTmp < PUMP_OFF_TEMP) digitalWrite (PUMP_PIN, LOW); // too cool for pump
   // writing a pin to the same state it was in already has no effect.  No state machine needed.
   battemp = sensors.getTempFByIndex(0); // we only have one
   sensors.requestTemperatures(); // get next one
   LastTempTime += TEMP_TIME; // set for next time
  } 
/*  if (LastPIDTime+PIDLOOPTIME <= millis())  // this construct works even when millis rolls over
  { // it's time to do speed control things
   LastPIDTime += PIDLOOPTIME; // set for next time, no slip
  // do the pid machine here
  
  FirstTime = 0; // Been there, done that
  }*/   
}
